/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file BasicPoincareMap2.h
///
/// @author Daniel Wilczak
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2008 by the CAPD Group.
//
// Distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

// Created on 29 April 2008, 00:24

#ifndef _BASICPOINCAREMAP2_H
#define	_BASICPOINCAREMAP2_H

#include <string>

namespace capd{
namespace poincare{
 
/**
 * Defines constants to specify section crossing direction
 * 
 * Section is given zeros of some scalar function.
 * PlusMinus means that we want to cross section comming from positive values 
 * and going to negative ones (MinusPlus means opposite direction).
 * Constant None means that we do not specify direction, both are allowed.
 */
enum CrossingDirection { PlusMinus = 1, None = 0, MinusPlus = -1};

/**
 *  BasicPoicareMap2 class is mainly used for non-rigorous computations of Poincare Map.
 *  For given Dynamical System and section it computes first return map
 *  and its derivatives but non-rigorously (i.e. without handling any computational errors). 
 *
 *  Let \f$ \varphi(t, x): R \times R^n -> R^n \f$ be a dynamical system
 *  Let \f$ s: R^n \to R \f$ be a section function,
 *  Let \f$ S = \{x \in R^n : s(x) = 0\} \f$
 *  Let \f$ P: S \to S \f$ be a Poincare Map
 *  For given point \f$ x \in S \f$ let T(x) be first return time (in given crossing direction)
 *  i.e.  \f$ P(x) = \varphi(T(x), x) \in S \f$
 *  In the following we denote by 
 *    - dP the derivative of Poincare Map P  : \f$ dP(x) = \frac{\partial P(x)}{\partial x} \f$
 *    - dT the derivative of T(x) : \f$ dT(x)  = \frac{\partial T(x)}{\partial x} \f$
 *    - dF the derivative of the flow : \f$ dF(x) = \frac{\partial \varphi(T(x), x)}{\partial x} \f$
 *  Then 
 *    \f$ dP = dF + \frac{\partial \varphi}{\partial t} dT \f$
 *
 *  Parameters:
 *   - DS    dynamical system
 *       - DS::MapType  vector field type
 *   - MapT  map describing section :
 *       - MapT::FunctionType - section function type
 *       - MapT               - gradient function type
 *       
 */  

template <typename DS, typename MapT = typename DS::MapType>
class BasicPoincareMap2
{
public:
  typedef DS DynamicalSystemType;                     ///< dynamical system type
  typedef typename DS::MapType VectorFieldType;       ///< vector field type
  typedef typename DS::MatrixType MatrixType;         ///< matrix type
  typedef typename DS::VectorType VectorType;         ///< vector type
  typedef typename DS::ScalarType ScalarType;         ///< scalar entries type
  
  typedef MapT MapType;                                ///< section gradient type
  typedef typename MapT::FunctionType FunctionType;    ///< Section function type
  
  
  typedef capd::poincare::CrossingDirection CrossingDirection;   
  
  /// Constructor
  BasicPoincareMap2( DynamicalSystemType  &ds,
             FunctionType & section,
             CrossingDirection direction = None, 
             double factor=0.1
        );
  
  /// Computes value of Poincare Map 
  VectorType operator()(VectorType v); 
  
  /// Computes value of Poincare Map and derivatives of the flow 
  VectorType operator()(VectorType v, MatrixType & dF);

  /// Computes Poincare Map and derivatives of the flow to given order
  template <typename CnCoeffType>
  CnCoeffType& operator()(CnCoeffType& result);

  // computations of derivatives of Poincare map from the derivatives of flow
  
  /// Computes dT from derivative of the flow
  VectorType computeDT(const VectorType& Px, const MatrixType& derivativeOfFlow);
  
  /// Computes derivative of Poincare Map dP 
  MatrixType computeDP(const VectorType& Px, const MatrixType& derivativeOfFlow, const VectorType& dT);
  
  /// Computes derivative of Poincare Map dP 
  MatrixType computeDP(const VectorType& Px, const MatrixType& derivativeOfFlow);
  
  /// Compute d2T/dx2 only for linear section
  MatrixType computeD2T(
           const VectorType& Px,
           const MatrixType& derivativeOfFlow,
           const VectorType& DT,
           const MatrixType& DP,
           MatrixType hessianOfFlow[]
         );
  /// Compute second derivative of Poincare Map d2P/dx2 only for linear section
  void computeD2P(
           const VectorType& Px,
           const MatrixType& derivativeOfFlow,
           const VectorType& DT,
           const MatrixType& DP,
           MatrixType hessianOfFlow[],
           const MatrixType& D2T,
           MatrixType result[]
      );

  /// Computes a partial derivatives of a Poincare map to an arbitrarily order only for linear section. 
  template<typename CnCoeffType>
  CnCoeffType computePartialDerivatives(const CnCoeffType& DPhi);

  DynamicalSystemType & getDynamicalSystem();              ///< Returns dynamical system used
  const DynamicalSystemType & getDynamicalSystem() const;  ///< Returns dynamical system used
  
  VectorFieldType & getField();               ///< Returns vector field
  const VectorFieldType & getField() const;   ///< Returns vector field

  FunctionType & getSection();                ///< Returns section function
  const FunctionType& getSection() const;     ///< Returns section function
  
  int getOrder() const;                       ///< Returns order of numerical method used
  const ScalarType& getStep() const;          ///< Returns time step              

  void setOrder(int newOrder);                ///< Sets order of Taylor method
  void setField(const MapType& newField);     ///< Sets vector field
  void setSection(const FunctionType& newSection);  ///< Sets new section function
  void setStep(const ScalarType& newStep);    ///< Sets time step for integration of DS

  void setFactor(double newFactor);           ///


protected:
  DS & dynamicalSystem;     ///< dynamical system (e.g. Taylor, C2Taylor or CnTaylor)
  FunctionType & section;   ///< section function 
  MapType gradient;         ///< gradient of section function
  CrossingDirection crossingDirection;  ///< wanted direction of section crossing
  
/// sectionFactor is used in the procedures reachSection and crossSection.
/// Before and after section crossing  we want the set to be closer to section 
/// than sectionFactor * sizeOfSet
  double sectionFactor;  
  int m_dim;                 ///< dimension of the space. 

}; // end of template PoincareMap2


// -----------------------  inline definitions ---------------------------------------------


template <typename DS, typename MapT>
inline BasicPoincareMap2<DS, MapT>::BasicPoincareMap2( 
   DynamicalSystemType  & ds,
   FunctionType & section,
   CrossingDirection direction,
   double factor
 ) : dynamicalSystem(ds), section(section), 
     crossingDirection(direction), sectionFactor(factor) {
  m_dim = dynamicalSystem.getField().getDimension();
  gradient = section.gradient(m_dim);
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline const typename BasicPoincareMap2<DS, MapT>::DynamicalSystemType & BasicPoincareMap2<DS, MapT>::getDynamicalSystem() const
{
  return dynamicalSystem;
}
// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline typename BasicPoincareMap2<DS, MapT>::DynamicalSystemType & BasicPoincareMap2<DS, MapT>::getDynamicalSystem() 
{
  return dynamicalSystem;
}
// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline typename BasicPoincareMap2<DS, MapT>::VectorFieldType const & BasicPoincareMap2<DS, MapT>::getField() const
{
  return dynamicalSystem.getField();
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline typename BasicPoincareMap2<DS, MapT>::VectorFieldType & BasicPoincareMap2<DS, MapT>::getField() 
{
  return dynamicalSystem.getField();
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline typename BasicPoincareMap2<DS, MapT>::FunctionType const & BasicPoincareMap2<DS, MapT>::getSection() const
{
  return section;
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline typename BasicPoincareMap2<DS, MapT>::FunctionType & BasicPoincareMap2<DS, MapT>::getSection() 
{
  return section;
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline int BasicPoincareMap2<DS, MapT>::getOrder() const
{
  return dynamicalSystem.getOrder();
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline const typename BasicPoincareMap2<DS, MapT>::ScalarType& BasicPoincareMap2<DS, MapT>::getStep() const
{
  return dynamicalSystem.getStep();
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline void BasicPoincareMap2<DS, MapT>::setField(const MapType& newField)
{
  dynamicalSystem.setField(newField);
  m_dim = newField.getDimension();
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline void BasicPoincareMap2<DS, MapT>::setOrder(int newOrder)
{
  dynamicalSystem.setOrder(newOrder);
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline void BasicPoincareMap2<DS, MapT>::setStep(const ScalarType& newStep)
{
  dynamicalSystem.setStep(newStep);
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline void BasicPoincareMap2<DS, MapT>::setFactor(double newFactor)
{
  sectionFactor = newFactor;
}

// -----------------------------------------------------------------------------------------

template <typename DS, typename MapT>
inline void BasicPoincareMap2<DS, MapT>::setSection(const FunctionType& newSection)
{
  
  section = newSection;
  gradient = section.gradient(m_dim);
}

/*  Expected interfaces of DS and MapT
 *
template <typename MapT>
class DynamicalSystemInterface{
  // Following internal types should be defined
  typedef MapT MapType;
  typedef typename MapType::MatrixType MatrixType;
  typedef typename MapType::VectorType VectorType;
  typedef typename MapType::ScalaType ScalarType;
  
  virtual MapType getField() = 0;
  void setField(MapType &) = 0;
  int  getOrder() = 0;
  void setOrder(int) = 0;
  ScalarType getStep() =0;
  void setStep(ScalarType) = 0;
  // NOT FINISHED
};


class SectionInterface{
  class FunctionType{
    virtual SectionInterface gradient() = 0;   
    
  };
  // NOT FINISHED
};*/

}} // namespace capd::poincare


#endif	/* _BASICPOINCAREMAP2_H */

/// @}
