
#ifndef _NBODYPROBLEM_H_
#define _NBODYPROBLEM_H_

#include <fstream>

#include "typedefs.h"
#include "nbody.h"
#include "textfile.h"
#include "mapping.h"

class NBodyProblem : public capd::newton::Mapping<IntervalMatrix>
{
  public:
    typedef IntervalMatrix::ScalarType IntervalType;
    typedef IntervalMatrix::RowVectorType VectorType;
    typedef IntervalMatrix MatrixType;
    typedef IntervalType::BoundType ScalarType;
    

    NBodyProblem(int A_nBodies, int A_nDiv, IntervalType A_time, 
                 int A_order, ScalarType A_step,
                 int A_Xorder, ScalarType A_Xstep, 
                 outputstream &A_str = scon);

    virtual ~NBodyProblem() {}

    IntervalMatrix operator[] (IntervalVector X);  // Derivative of the Map
    IntervalVector operator() (IntervalVector X);  // Value of the Map
    IntervalVector operator() (IntervalVector X, IntervalMatrix &dF);  // Value of the Map and its derivative dF

    int getDimension() const;
    bool uniquenessTest();


    int nBodies;
    int nDiv;
    IntervalVector calculatePM(IntervalVector X, IntervalMatrix &dP);
    IntervalMatrix monodromyMatrix(IntervalVector X);
    IntervalMatrix calculateDiff(const IntervalVector &X);

   protected:
    Interval time;
    int order,
        Xorder;
    ScalarType step,
               Xstep;
    Map vectorField;
    Taylor taylor;
    TimeMap timeMap;
    IntervalMatrix  E,
                    Shift,
                    R;
    IntervalVector setZC;        // Set Z x C for uniqueness test

    outputstream & log;

    IntervalMatrix generateE();
    IntervalMatrix generateShift();
    IntervalMatrix generateR();
    IntervalMatrix generateBodiesShift();
    IntervalMatrix generateBodiesShift_1();


};



inline NBodyProblem::NBodyProblem(int A_nBodies, int A_nDiv, IntervalType A_time,
                           int A_order, ScalarType A_step, 
                           int A_Xorder, ScalarType A_Xstep,  
                           outputstream &A_str)
                           : nBodies(A_nBodies),
                             nDiv(A_nDiv),
                             time(A_time),
                             order(A_order),
                             Xorder(A_Xorder),
                             step(A_step),
                             Xstep(A_Xstep), 
                             vectorField(reducedNBodyEquation2D(A_nBodies)),
                             taylor(vectorField,A_order, A_step), 
                             timeMap(taylor),
                             log(A_str)
{
   E = generateE();
   Shift = generateShift();
   R = generateR();
}



inline int NBodyProblem::getDimension() const
{
   return ((nBodies-1)*4*nDiv-2);
}

// Generates matrix of expansion function
// simply, this function adds two coordinates set to zero
//   (second and third coordinate)
//

inline IntervalMatrix NBodyProblem::generateE()
{
   IntervalMatrix E((nBodies-1)*4*nDiv,(nBodies-1)*4*nDiv-2);
   E.clear();
   E[0][0]=Interval(1.);
   for(int i=1; i<(nBodies-1)*4*nDiv-2; ++i)
      E[i+2][i]=Interval(1.);
   return E;
}


inline IntervalMatrix NBodyProblem::generateR()
{
   IntervalMatrix R((nBodies-1)*4*nDiv-2,(nBodies-1)*4*nDiv);
   R.clear();
   R[0][0]=Interval(1.);
   for(int i=1; i<(nBodies-1)*4*nDiv-2; i++)
      R[i][i+2]=Interval(1.);
   return R;
}

inline IntervalMatrix NBodyProblem::generateShift()
{
   IntervalMatrix S((nBodies-1)*4*nDiv,(nBodies-1)*4*nDiv);
   S.clear();
   int tupleLength = (nBodies-1)*4;
   for(int i=(nDiv-1)*tupleLength, j=0; i< nDiv*tupleLength; i++, j=(j+1)%4)
     S[j][i] = Interval(-1.);

   for(int i=(nDiv-1)*tupleLength, j=4; i< nDiv*tupleLength-4; i++, j++)
     S[j][i] = Interval(1.);

   for(int i=0; i< (nDiv-1)*tupleLength; i++)
     S[tupleLength+i][i] = Interval(1.);

   return S;
}

#endif
