#include "nBodyProblem2.h"
#include "subVector.h"
#include <exception>

// Dimension of the phase space
#define  DIM_SP (4*nBodies-4)
#define  DIM_NORMAL (4*nBodies)

// Dimension of full space in parallel shooting
#define  DIM_FULL  (DIM_SP * nDiv)

// Dimension of reduced space
#define  DIM_RED (DIM_FULL-2)


IntervalMatrix NBodyProblem2::operator[] (IntervalVector X)
{

    IntervalVector X_e = E * X;

    IntervalMatrix dP(DIM_FULL, DIM_FULL);

    setZC = Shift * calculatePM(X_e, dP, XorderBegin, XorderEnd, XorderStep,
                                XstepBegin, XstepEnd, XstepFactor);        //  Compute Poicare Map and its derivative

    IntervalMatrix dF = (R * Shift * dP * E - IntervalMatrix::Identity(DIM_RED));

    return  dF;

}

IntervalVector NBodyProblem2::operator() (IntervalVector X)
{

    IntervalVector X_e = E * X;

    IntervalMatrix dP(DIM_FULL, DIM_FULL);

    IntervalVector PX = calculatePM(X, dP, orderBegin, orderEnd, orderStep,
                                    stepBegin, stepEnd, stepFactor);        //  Compute Poicare Map and its derivative

    IntervalVector F = (R * Shift * PX  - X);

    return  F;
}

IntervalVector NBodyProblem2::operator() (IntervalVector X, IntervalMatrix &dF)
{

    IntervalVector X_e = E * X;

    IntervalMatrix dP(DIM_FULL, DIM_FULL);

              //  Compute Poicare Map and its derivative
    IntervalVector PX = calculatePM(X_e, dP, orderBegin, orderEnd, orderStep,
                                    stepBegin, stepEnd, stepFactor);


    dF = (R * Shift * dP * E - IntervalMatrix::Identity(DIM_RED));

    IntervalVector F = (R * Shift * PX  - X);

    return  F;

}


/////////////////////////////////////////////////////////////////////////////////
///  Calculates Poincare Map and its derivative with diferent time steps
////////////////////////////////////////////////////////////////////////////////

IntervalVector NBodyProblem2::calculatePM(IntervalVector X, IntervalMatrix &dP,
                 int orderBegin, int orderEnd, int orderStep, ScalarType stepBegin,
                 ScalarType stepEnd, ScalarType stepFactor)
{

   IntervalVector x0(DIM_SP),
                  tempResult(DIM_SP);
   IntervalVector result(DIM_FULL);
   IntervalMatrix derivative(DIM_FULL, DIM_FULL),
                  dTemp(DIM_SP, DIM_SP);
   LNorm norm;
   derivative.clear();


   for(int d=0; d < nDiv; d++)
   {
      scon << "\n part : " << d;
      bool firstTime = true;
      for(int order = orderBegin; order<=orderEnd; order+=orderStep)
      {
         timeMap.setOrder(order);

         for(ScalarType step = stepBegin; step <=stepEnd; step*=stepFactor)
         {

            scon << "\n order = " << order << "    step = " << step;

            timeMap.setStep(step);

            for(int i=0; i<DIM_SP; i++)
               x0[i]=X[DIM_SP*d+i];

            try
            {
              C1rect2 S(x0,norm);

              tempResult = timeMap(time, S);
              dTemp = IntervalMatrix(S);

              if(firstTime)
              {
                 copyToSubvector(tempResult, result, d*DIM_SP, (d+1)*DIM_SP);
                 copyToSubmatrix(dTemp, derivative, d*DIM_SP, (d+1)*DIM_SP);
                 firstTime = false;
              }
              else
              {
                 intersectionWithSubvector(tempResult, result,  d*DIM_SP, (d+1)*DIM_SP);
                 intersectionWithSubmatrix(dTemp, derivative,  d*DIM_SP, (d+1)*DIM_SP);
              }

            }
            catch(std::exception &e)
            { }

            scon << " vector size = " << size(tempResult) << " matrix[0] size " << size(dTemp[0]);


         }
      }
      if(firstTime)
         throw std::runtime_error(" Error calculating Poincare Map. " );
   }

   dP = derivative;
   return result;
}

bool NBodyProblem2::uniquenessTest()
{
   IntervalVector C(2), c0(2);
   IntervalVector setZC_1(DIM_SP);

   // We take only entries from the first configuration
   for(int i=0; i<DIM_SP; ++i)
      setZC_1[i] = setZC[i];

   C[0] = setZC[1];  C[1] = setZC[2];
   c0[0] = 0.0;      c0[1] = 0.0;

   Map J = reducedFirstIntegrals2D(nBodies);

   IntervalMatrix dJ_dx = J[setZC_1];

   IntervalMatrix dJ_dc(2,2);

   for(int i=0; i<2; i++)
      for(int j=1; j<3; j++)
         dJ_dc[i][j-1] = dJ_dx[i][j];

   Interval det = dJ_dc[0][0]*dJ_dc[1][1] - dJ_dc[0][1]*dJ_dc[1][0];

   sout << "\n\n C = " << C
        << "\n\n dJ/dc(Z,C) = \n" << dJ_dc
        << "\n\n det(dJ/dc(Z,C)) = " << det;

   if((det.contains(0.0))||(!subset(c0,C)))
      return false;
   else
      return true;

}
