// This is TEXTFILE.H. Copyright (C) 1997-2004 by Pawel Pilarczyk.
// It contains classes 'textfile', 'outputstream' and other useful things.
// This is free software. No warranty. Consult 'license.txt' for details.
// Started in July 1997. Last revision: October 10, 2004.


#ifndef _TEXTFILE_H_
#define _TEXTFILE_H_

#include "options.h"

#include <ctime>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>


// classes defined within this header file:
class textfile;
class outputstream;


// --------------------------------------------------
// ----------------- OUTPUT STREAM ------------------
// --------------------------------------------------

class outputstream
{
public:
	// default constructor and destructor
	outputstream (std::ostream &_out = std::cout,
		bool _show = true, bool _flushit = false);
	~outputstream ();

	// turn on logging to a file
	void logfile (const char *filename);

	// turn on logging to the same file as the other stream
	void logfile (const outputstream &other);

	// should this output be shown to 'cout'?
	bool show;

	// should this output be logged to the file?
	bool log;

	// should this output be flashed every time?
	bool flushit;

	// retrieve the log stream
	std::ofstream *getlogstream (void);

	// the main output stream
	std::ostream &out;

	// keep the allocated file open for ever (don't close it)
	void keepforever (void);

protected:

	// log file
	std::ofstream *logstream;

	// was this log file copied from another output stream?
	bool copied;

}; /* class outputstream */
//End//

// --------------------------------------------------

inline outputstream::outputstream (std::ostream &_out, bool _show,
	bool _flushit): out (_out)
{
	show = _show;
	flushit = _flushit;
	log = false;
	logstream = NULL;
	copied = false;
	return;
} /* outputstream::outputstream */

inline void outputstream::logfile (const char *filename)
{
	if (logstream && !copied)
		delete logstream;
	copied = false;
	logstream = new std::ofstream (filename);
	if (!logstream || !*logstream)
	{
		out << "WARNING: Can't create '" << filename <<
			"'. Logging to a file has been turned off." <<
			std::endl;
		if (logstream)
			delete logstream;
		logstream = NULL;
	}
	else
		log = true;
	return;
} /* outputstream::logfile */

inline void outputstream::logfile (const outputstream &other)
{
	if (!copied && logstream)
		delete logstream;
	logstream = other. logstream;
	if (logstream)
	{
		copied = true;
		log = true;
	}
	return;
} /* outputstream::logfile */

inline outputstream::~outputstream ()
{
	if (!copied && logstream)
		delete logstream;
	return;
} /* outputstream::~outputstream */

inline std::ofstream *outputstream::getlogstream (void)
{
	return logstream;
} /* outputstream::getlogstream */

inline void outputstream::keepforever (void)
{
	copied = true;
	return;
} /* outputstream::getlogstream */

//Doc//
template<class type>
inline outputstream &operator << (outputstream &out, const type &object)
//End//
{
	if (out. flushit)
	{
		if (out. show)
			out. out << object << std::flush;
		if (out. log && out. getlogstream ())
			(*(out. getlogstream ())) << object << std::flush;
	}
	else
	{
		if (out. show)
			out. out << object;
		if (out. log && out. getlogstream ())
			(*(out. getlogstream ())) << object;
	}
	return out;
} /* operator << */

//Doc//
inline outputstream &operator << (outputstream &out, const char *object)
//End//
{
	if (out. flushit)
	{
		if (out. show)
			out. out << object << std::flush;
		if (out. log && out. getlogstream ())
			(*(out. getlogstream ())) << object << std::flush;
	}
	else
	{
		if (out. show)
			out. out << object;
		if (out. log && out. getlogstream ())
			(*(out. getlogstream ())) << object;
	}
	return out;
} /* operator << */

//Doc//
// --------------------------------------------------

// console output with all the junk that spoils the log file
extern outputstream scon;
// standard output, optionally with logging
extern outputstream sout;
// messages only for logging, ignored if the log file is not in use
extern outputstream slog;
// additional debug messages
extern outputstream sbug;
// a stream to keep sequences of processed data
extern outputstream sseq;


// --------------------------------------------------
// ----------------- various tools ------------------
// --------------------------------------------------

template <class type>
inline void swapelements (type &x, type &y)
//End//
{
	type z = x;
	x = y;
	y = z;
	return;
} /* swapelements */

//Doc//
template <class type>
inline int sortelements (type *tab, int n)
// Sort elements, remove repeated elements and return their number.
//End//
{
	switch (n)
	{
		case 0:
			return 0;
		case 1:
			return 1;
		case 2:
			if (tab [0] == tab [1])
				return 1;
			else if (tab [0] > tab [1])
				swapelements (tab [0], tab [1]);
			return 2;
		default:
			for (int i = 0; i < n - 1; i ++)
			{
				for (int j = i + 1; j < n; j ++)
					if (tab [i] > tab [j])
						swapelements (tab [i],
							tab [j]);

				if (tab [i] == tab [i + 1])
				{
					n --;
					for (int j = i + 1; j < n; j ++)
						tab [j] = tab [j + 1];
				}
			}
			break;
	}
	return n;
} /* sortelements */

//Doc//
// a simple procedure for ignoring the input until the end of a line
inline void ignoreline (std::istream &in)
//End//
{
	in. ignore (0x7FFFFFFFl, '\n');
	return;
} /* ignoreline */

//Doc//
// a simple procedure for ignoring white characters (spaces, tabulators,
// CRs and LFs), as well as comments from the input text file
inline void ignorecomments (std::istream &in)
//End//
{
	int ch = in. peek ();
	while (true)
	{
		switch (ch)
		{
		case ';':
			ignoreline (in);
			ch = in. peek ();
			break;
		case ' ':
		case '\t':
		case '\r':
		case '\n':
			in. get ();
			ch = in. peek ();
			break;
		default:
			return;
		}
	}	
} /* ignorecomments */

//Doc//
// get the matching parenthesis for the given one or EOF if none
inline int closingparenthesis (int ch)
//End//
{
	switch (ch)
	{
		case '(':
			return ')';
		case '{':
			return '}';
		case '[':
			return ']';
		case '<':
			return '>';
		case '/':
			return '/';
		default:
			return EOF;
	}
} /* closingparenthesis */

//Doc//
// a simple procedure for reading an opening parenthesis from the input file;
// return a corresponding closing parenthesis or EOF if none was found
inline int readparenthesis (std::istream &in)
//End//
{
	int closing = closingparenthesis (in. peek ());
	if (closing != EOF)
		in. get ();
	return closing;
} /* readparenthesis */

//Doc//
// retrieve the command line as a single string
inline std::string commandline (int argc, char *argv [])
//End//
{
	std::string s;
	for (int i = 0; i < argc; i ++)
	{
		if (i)
			s += ' ';
		s += argv [i];
	}
	return s;
} /* commandline */

//Doc//
// set the given extension to the file name (replace its extension if any)
const char *setextension (const char *filename, const char *ext);

// retrieve the current time
inline const char *currenttime (void)
//End//
{
	time_t t;
	time (&t);
	return asctime (localtime (&t));
} /* currenttime */

//Doc//
// Write the given set to a file whose name is a concatenation of the prefix
// and the given file name unless the prefix is NULL.
// Display a warning only if unsuccessful.
// This procedure can be applied to any data type which has the operator <<
// for writing it to a text output stream (ostream) defined.
template <class settype>
static void savetheset (const settype &c, const char *prefix,
	const char *filename, const char *name)
//End//
{
	// if there is no prefix given, do not save the file
	if (!prefix)
		return;

	// prepare the full file name
	std::string str;
	if (prefix)
		str = std::string (prefix) + std::string (filename);
	else
		str = filename;
	const char *s = str. c_str ();

	// create the output file
	std::ofstream out (s);
	if (!out)
	{
		sout << "WARNING: Cannot save the file '" << s << "'.\n";
		return;
	}

	// write the cubes to the output file
	sout << "Saving " << name << " to '" << s << "'... ";
	out << "; This is " << name << ".\n";
	out << c;
	sout << "Done.\n";

	return;
} /* savetheset */


//Doc//
// --------------------------------------------------
// ----------------- ERROR MESSAGE ------------------
// --------------------------------------------------
// If you want to throw an error message, you type something like
// the following:
//
//	ERRORMSG << "Wrong number: " << n << " instead of 0." << THROW
//
// Note that there is no semicolon at the end. This is important
// especially in the setting like below:
//
//	if (n != 0)
//		ERRORMSG << .......... << THROW
//	else
//		..........
//
// This may probably not work properly, but at the moment I couldn't come up
// with anything better.

#define ERRORMSG {std::ostringstream error_msg_string; error_msg_string
#define THROW ""; throw error_msg_string. str (). c_str (); }

// --------------------------------------------------

// throw a message about the inability to do something with a file
inline void fileerror (const char *filename, const char *what = "open")
//End//
{
	ERRORMSG << "Cannot " << what << " the file '" <<
		filename << "'." << THROW
	return;
} /* fileerror */


//Doc//
// --------------------------------------------------
// -------------------- TEXTFILE --------------------
// --------------------------------------------------
// A file for reading text data.
// All blank characters and comments are ignored.

// the maximal length of a phrase read at a time
#define MAXTEXTLENGTH 200

class textfile
{
public:
	// default constructor
	textfile ();

	// a constructor for an opened file
	textfile (std::istream &in);

	// a constructor which opens a file
	//	textfile (char *filename);

		// open the file; return: 1 = already opened, 0 = ok,
		// -1 = an error occurred (no message is displayed)
		int open (const char *filename);

		// close the file
		void close ();

		// default destructor
		~textfile ();

		// read an integer number;
		// if none can be read, return the suggested default value;
		// if requested, a preceding ':', '=' or ',' may be ignored
		long readnumber (long defaultnumber = 0,
			int ignorecolons = 0);

		// check the first non-white character waiting to be read;
		// return EOF if end-of-file has been reached
		int checkchar ();

		// read one non-white character from the file;
		// return EOF if end-of-file has been reached
		int readchar ();

		// read one word from the file and return its actual length
		int readword (char *buf, int maxlength);

		// read a strongly expected phrase from the file (ignore
		// white characters before the phrase and inside it);
		// return: 0 = ok, -1 = not found (show an error message)
		int readphrase (const char *phrase);

		// ignore the current line
		void ignoreline ();

		// return the current line number
		long linenumber ();

		// return the text file's status: 0 = not opened,
		// 1 = opened, 13 = initialized with an input stream
		operator int ();


	protected:

		// the character already read and waiting to be interpreted
		int ch;

		// current line number
		long line;

		// the true file if it was opened
		std::ifstream *fs;

		// a pointer to an input stream (usually equal to 'fs')
		std::istream *s;

		// ignore white characters and fill in the variable 'ch'
		// with the first non-white character
		void ignorespaces ();

}; /* class textfile */
//End//

#endif // _TEXTFILE_H_

