/////////////////////////////////////////////////////////////////////////////
//   Package:          Choreographies in the N-Body Problem
/// @addtogroup NBodyProblem
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file nBodyProblem.h
///
/// Definition of the class NBodyProblem2 
///
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////
// Copyright (C) Tomasz Kapela 2006. 

#ifndef _NBODYPROBLEM_H_
#define _NBODYPROBLEM_H_

#include "typedefs.h"
#include "nbody.h"
#include "textfile.h"
#include "mapping.h"

//////////////////////////////////////////////////////////////////////////////
//   NBodyProblem
///
///   NBodyProblem class is used for choreography existence proofs. 
///   NBodyProblem can be seen as a function whos zeroes are initial conditions
///   of some choreography. 
///   It is composition of four function
///       \f[ R \circ \sigma \circ P \circ E \f] 
///   where <br>
///   <ul>
///       <li> E - expands reduced initional conditions to the phase space dimension </li>
///       <li> P - is Time Map of Nbody Equation flow </li>
///       <li> \f$\sigma\f$ - is cyclical shift of particles</li>
///       <li> R - reduces configurations </li>
///   </ul>   
///
///   @author Tomasz Kapela     @date 22-08-2006
//////////////////////////////////////////////////////////////////////////////


class NBodyProblem : public capd::newton::Mapping<IntervalMatrix>
{
  public:  
    
    typedef IntervalMatrix::ScalarType IntervalType;
    typedef IntervalMatrix::RowVectorType VectorType;
    typedef IntervalMatrix MatrixType;
    typedef IntervalType::BoundType ScalarType;
    
    NBodyProblem(int A_nBodies, int A_nDiv, Interval A_time, 
                 int A_order, ScalarType A_step, 
                 int A_Xorder, ScalarType A_Xstep);

    virtual ~NBodyProblem() {}

    MatrixType operator[] (VectorType X);                      ///< Derivative of the Map
    VectorType operator() (VectorType X);                      ///< Value of the Map
    VectorType operator() (VectorType X, MatrixType& dF);      ///< Value of the Map and its derivative dF

    int dimension();
    bool uniquenessTest();

    VectorType calculatePM(VectorType X, MatrixType& dP);
    MatrixType monodromyMatrix(const VectorType& X);
    MatrixType calculateDiff(const VectorType& X);

  protected:
    MatrixType generateE();
    MatrixType generateShift();
    MatrixType generateR();
    MatrixType generateBodiesShift();
    MatrixType generateBodiesShift_1();

  public:
    int nBodies;               ///< number of bodies
    int nDiv;                  ///< number of divisions in parellel shooting  
    
  protected:
    Interval time;             ///< we go with the flow for a time 'time' 
    int order,                 ///< order of Taylor method in computation for a point
        Xorder;                ///< order of Taylor method in computation for a set 
    ScalarType step,           ///< time step of Taylor method in computation for a point
           Xstep;              ///< time step in computation for a set 
    Map vectorField;           ///< vector field of reduced NBody equation
    Taylor dynamicalSystem;    ///< dynamical system generated by NBody equation
    TimeMap timeMap;           ///< class that computes Time Map and its derivative
    MatrixType      E,         ///< matrix of a linear expansion map
                    Shift,     ///< matrix of a shift map
                    R;         ///< matrix of a linear reduction map
    VectorType     setZC;      ///< Set Z x C for uniqueness test
};



inline NBodyProblem::NBodyProblem(int A_nBodies, int A_nDiv, Interval A_time,
                           int A_order,  ScalarType A_step,
                           int A_Xorder, ScalarType A_Xstep)
                           : nBodies(A_nBodies),
                             nDiv(A_nDiv),
                             time(A_time),
                             order(A_order),
                             Xorder(A_Xorder),
                             step(A_step),
                             Xstep(A_Xstep),
                             vectorField(reducedNBodyEquation2D(A_nBodies)),
                             dynamicalSystem(vectorField,A_order, A_step),
                             timeMap(dynamicalSystem)
{
   E = generateE();
   Shift = generateShift();
   R = generateR();
}



inline int NBodyProblem::dimension()
{
   return ((nBodies-1)*4*nDiv-2);
}

#endif

/// @}
